/*
** filterlib.c - written in milano by vesely on 19 oct 2001
** sophos anti virus for courier-mta global filters (with
** variations from original libfilter.c)
**
** This is a modular software, not object oriented:
** compile with:
**       "-DFILTER_NAME=blah_blah"
*/
/*
** Copyright (c) 1999, 2000, 2001, 2002, 2003, 2004 Alessandro Vesely
** All rights reserved. see COPYING
*/
#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>

/* COURIER */
#include "unconfig.h"
#include <courier/filtersocketdir.h>
#include <libfilter/libfilter.h>

#if defined(HAVE_CONFIG_H)
#include "unconfig.h"
#include <config.h>
#endif

#if defined(FILTER_NAME)
#define STRING2(P) #P
#define STRING(P) STRING2(P)
#define THE_FILTER STRING(FILTER_NAME)
#else
#define THE_FILTER "filterlib"
#endif

#include "filterlib.h"

static volatile int
	signal_timed_out = 0,
	signal_break = 0,
	signal_hangup = 0;
static int live_children = 0;

static char const filter_pass[] = "X-checked-" THE_FILTER;

typedef struct ctl_fname_chain
{
	 struct ctl_fname_chain *next;
	 char fname[1]; // actually large as needed
} ctl_fname_chain;

struct filter_lib_struct
{
	void *parm;
	unsigned int verbose:4;
	unsigned int testing:4;
	unsigned int batch_test:4;
	fl_callback filter_fn;

	char const *resp;
	int in, out;
	char *data_fname;
	FILE *data_fp;
	ctl_fname_chain *cfc;
	char *argv0;
	int ctl_count;
};

/* ----- sig handlers ----- */

static int sig_verbose = 0;
static void child_reaper(int sig)
{
	int status;
	pid_t child;

	(void)sig;
	
	while ((child = waitpid(-1, &status, WNOHANG)) > 0)
	{
		--live_children;
#if !defined(NDEBUG)
		if (sig_verbose > 0)
		{
			char buf[80];
			unsigned s = snprintf(buf, sizeof buf,
				THE_FILTER ": %d exited, %d remaining\n",
				(int)child, live_children);
			if (s >= sizeof buf)
			{
				buf[sizeof buf - 1] = '\n';
				s = sizeof buf;
			}
			write(2, buf, s);
		}
#endif
	}
}

static void sig_catcher(int sig)
{
#if !defined(NDEBUG)
	if (sig_verbose > 0)
	{
		char buf[80];
		unsigned s = snprintf(buf, sizeof buf,
			THE_FILTER "[%d]: received signal %d: keep=%d\n",
			(int)getpid(), sig, fl_keep_running());
		if (s >= sizeof buf)
		{
			buf[sizeof buf - 1] = '\n';
			s = sizeof buf;
		}
		write(2, buf, s);
	}
#endif
	switch(sig)
	{
		case SIGALRM:
			signal_timed_out = 1;
			break;

		case SIGHUP:
		case SIGUSR1:
		case SIGUSR2:
			signal_hangup = sig;
			break;
		
		case SIGPIPE:
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			signal_break = 1;
			break;
		default:
			break;
	}
}

void fl_alarm(unsigned seconds)
{
#if !defined(NDEBUG)
	/* check the signal handler has been set up */
	struct sigaction oact;
	if (sigaction(SIGALRM, NULL, &oact) != 0 ||
		oact.sa_handler != &sig_catcher)
			fprintf(stderr, "SIGALRM not correct!!\n");
#endif	
	signal_timed_out = 0;
	alarm(seconds);
}

/* ----- ctl_fname_chain handling ----- */

static ctl_fname_chain* cfc_shift(ctl_fname_chain **cfc)
{
	ctl_fname_chain *s = *cfc;
	if (s)
		*cfc = s->next;
	return s; // freed by caller
}

static int cfc_unshift(ctl_fname_chain **cfc, char const *fname, size_t len)
{
	ctl_fname_chain *const u = (ctl_fname_chain*)
		malloc(len + sizeof(ctl_fname_chain));
	if (u)
	{
		u->next = *cfc;
		memcpy(u->fname, fname, len);
		u->fname[len] = 0;
		*cfc = u;
		return 0;
	}
	return 1;
}

/* ----- fl_* aux functions ----- */

void *fl_get_parm(fl_parm*fl)
{
	return fl->parm;
}

void fl_set_parm(fl_parm *fl, void* parm)
{
	fl->parm = parm;
}

void fl_set_verbose(fl_parm*fl, int verbose)
{
	fl->verbose = sig_verbose = verbose;
}

int fl_get_verbose(fl_parm*fl)
{
	return fl->verbose;
}

int fl_keep_running(void)
{
	return signal_timed_out == 0 && signal_break == 0;
}

void fl_pass_message(fl_parm*fl, char const *resp)
{
	fl->resp = resp;
}

FILE* fl_get_file(fl_parm*fl)
{
	return fl->data_fp;
}

fl_test_mode fl_get_test_mode(fl_parm* fl)
{
	return fl->testing ? fl->batch_test ?
		fl_batch_test : fl_testing : fl_no_test;
}

/* ----- drop message ----- */

static int count_recipients(FILE *fp, char **from_mta)
{
	int count = 0;
	char buf[2048];
	
	if (fseek(fp, 0, SEEK_SET) != 0)
		return -1;

	while (fgets(buf, sizeof(buf), fp))
	{
		char *ends = strchr(buf, '\n');
		if (ends)
			*ends = 0;
		else
		{
			int c;
			while ((c = fgetc(fp)) != EOF && c != '\n')
				continue;
		}
		
		if (buf[0] == 'r')
			++count;
		else if (buf[0] == 'f' && from_mta && *from_mta == NULL)
			*from_mta = strdup(&buf[1]);
		
	}
	if (ferror(fp) || fseek(fp, 0, SEEK_END) != 0)
		return -1;

	return count;
}

int fl_drop_message(fl_parm*fl, char const *reason)
{
	int ctl = 0, rtc = 0;
	time_t tt;
	char *from_mta = NULL;
	
	if (fl->verbose > 4)
	{
		fprintf(stderr,
			THE_FILTER "[%d]: about to drop msg with %d ctl file(s)\n",
			(int)getpid(), fl->ctl_count);
	}

	if (fl_get_test_mode(fl) == fl_testing)
		return 1; // dummy ctlfile given (the ids)
	
	time(&tt);

	while (fl->cfc)
	{
		int irtc = 0, goterrno = 0;
		ctl_fname_chain *cfc = cfc_shift(&fl->cfc);
		FILE *fp;
		
		++ctl;
		errno = 0;
		fp = fopen(cfc->fname, "r+");
		if (fp)
		{
			int i;
			int const count = count_recipients(fp, &from_mta);

			/*
			** the ugly hack: since submit writes various records
			** ("8", "U", "V", "w", "E", "p" "W" and "A") _after_
			** running global filters, we have to estimate enough
			** space so that it won't overwrite our faked delivery
			**
			** we put 254, about 32 bytes per record is enough and
			** we still allowing any unlikely old-fashioned fgets
			** with 256 bytes of buffer
			*/
			fprintf(fp, "%254s\n", "");
			
			for (i = 0; i < count && !ferror(fp); ++i)
				fprintf(fp, "I%d R 200 Dropped.\nS%d %ld\n",
					i, i, (long)tt);
			fprintf(fp, "C%ld\n", (long)tt);
			
			if (count < 0 || ferror(fp))
			{
				irtc = 1;
				if (goterrno == 0)
					goterrno = errno;
			}
			fclose(fp);
			if (fl->verbose > 4)
			{
				fprintf(stderr,
					THE_FILTER
					"[%d]: dropped %d/%d recipient(s) time=%ld on ctl file %s\n",
					(int)getpid(), i, count, (long)tt, cfc->fname);
			}
		}
		else
		{
			irtc = 1;
			goterrno = errno;
		}
		
		if (irtc)
		{
			rtc = irtc;
			fprintf(stderr, "ALERT:"
				THE_FILTER "[%d]: error on ctl file %d/%d (%s): %s\n",
				(int)getpid(), ctl, fl->ctl_count, cfc->fname, strerror(goterrno));
		}
		else if (fl->verbose)
		/*
		** main logging function
		** (given as error as the rest of refused messages)
		*/
		{
			fprintf(stderr,
				"ERR:dropped,From-MTA=<%s>: "
				THE_FILTER "[%d]: %s",
				from_mta ? from_mta : "",
				(int)getpid(),
				reason ? reason : "w/o apparent reason\n");
		}
		
		free(cfc);
	}
	
	free(from_mta);
	return rtc;
}

#if 0
int fl_is_pass_signed(fl_parm*fl, char const *hdr, size_t length)
{
	if (length > sizeof filter_pass &&
		strncmp(filter_pass, hdr, sizeof filter_pass - 1) == 0)
	{
		char const *text = hdr + sizeof filter_pass;
		char const *end = hdr + length;
		//while (text < end && ...)
	}
	return 0;
}
#endif

/* ----- core filter functions ----- */

static int read_fname(fl_parm* fl)
/*
** This reads data and ctl file names. A file name must be either
** an absolute path or relative to the current directory (as usual.)
** When runnining as an installed filter, lf_init does chdir to
** COURIER_HOME, which is what was configured as @prefix@: so this
** is consistent whith courierfilter.html which states that path
** names may be relative to that directory. (However, it looks as
** if "LOCALSTATEDIR/tmp" is always used as a an absolute path.)
*/
{
	char buf[1024];
	unsigned count = 0;
	int found = 0, empty = 0, rtc = 0;
	int const fd = fl->in;
	fl->data_fname = NULL;

#if !defined(NDEBUG)
	if (fl->verbose > 1)
		fprintf(stderr, THE_FILTER "[%d]: reading fd %d\n",
			(int)getpid(), fd);
#endif
	
	fl_alarm(30);
	while (count < sizeof buf && empty == 0 && fl_keep_running())
	{
		unsigned const p = read(fd, &buf[count], 1);
		if (p == (unsigned)(-1))
		{
			switch (errno)
			{
				case EAGAIN:
				case EINTR:
					continue;
				default:
					break;
			}
			fprintf(stderr, "ALERT:"
				THE_FILTER "[%d]: cannot read fname pipe: %s\n",
				(int)getpid(), strerror(errno));
			break;
		}
		else if (p != 1)
		{
			fprintf(stderr, "ALERT:"
				THE_FILTER "[%d]: Unexpected %s (%d) on fname pipe\n",
				(int)getpid(), p == 0 ? "EOF" : "rtc from read", p);
			break;
		}

		if (buf[count] != '\n')
			++count;
		else if (count == 0) /* empty line ends filenames */
			empty = 1;
		else
		{
			buf[count] = 0;

#if !defined(NDEBUG)
			if (fl->verbose > 1)
				fprintf(stderr, THE_FILTER 
					"[%d]: piped fname[%d]: %s (len=%u)\n",
					(int)getpid(), found, buf, count);
#endif

			if (++found == 1) /* first line */
				fl->data_fname = strdup(buf);
			else if (cfc_unshift(&fl->cfc, buf, count))
				fprintf(stderr, "ALERT:"
					THE_FILTER "[%d]: malloc on filename #%d\n",
					(int)getpid(), found);
			count = 0;
		}
	}
	
	alarm(0);
	fl->ctl_count = found - 1;
	if (!fl_keep_running() || found < 2 ||
		empty == 0 || count >= sizeof buf || fl->data_fname == NULL)
	{
		rtc = 1;
		if (!fl_keep_running())
			fprintf(stderr, "ALERT:"
				THE_FILTER
				"[%d]: reading fname pipe terminated prematurely\n",
					(int)getpid());
		if (found != 2 || empty == 0 || fl->data_fname == NULL)
			fprintf(stderr, "ALERT:"
				THE_FILTER "[%d]: found%s %d file name(s)%s%s\n",
				(int)getpid(),
				found < 2 ? " only" : "", found,
				empty ? "" : " no empty line",
				fl->data_fname == NULL ? " malloc failure" : "");
		if (count >= sizeof buf)
			fprintf(stderr, "ALERT:"
				THE_FILTER "[%d]: Buffer overflow reading fname pipe\n",
					(int)getpid());
		free(fl->data_fname);
		fl->data_fname = NULL;
	}

	return rtc;
}

#if !defined(NDEBUG)
static void fl_break(void)
{
	fputs("execution resumed\n", stderr);
}
#endif

static void fl_reset_signal(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof act);

	sigemptyset(&act.sa_mask);
	act.sa_handler = SIG_DFL;

	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGHUP, &act, NULL);
	sigaction(SIGUSR1, &act, NULL);
	sigaction(SIGUSR2, &act, NULL);
}

static void fl_runchild(fl_parm* fl)
{
	pid_t pid;
	int retry = 2;
	int const fd_out = fl->out;
	
	while ((pid = fork()) < 0 && --retry >= 0)
		sleep(1);
	
	if (pid == 0) /* child */
	{
		char const *resp = NULL;
		unsigned w = 0, l;
		int rtc = 0;
		
		if (fl->verbose > 3)
			fprintf(stderr, THE_FILTER "[%d]: started child\n",
				(int)getpid());
			
#if !defined(NDEBUG)
		if (fl_get_test_mode(fl) == fl_testing)
		{
			char *debug = getenv("DEBUG_FILTER");
			if (debug)
			{
				int nsec = atoi(debug);
				if (nsec <= 0) nsec = 30;
				
				/* hallo! */
				fprintf(stderr, "DEBUG_FILTER: %s\n"
					"now you have %d secs to\n"
					"%% gdb %s %d\n"
					"(gdb) break fl_break\n"
					"Breakpoint 1 at 0xxx: file %s, line %d.\n"
					"(gdb) cont\n",
						debug, nsec, fl->argv0, (int)getpid(),
						__FILE__, __LINE__ -40);
				nsec = sleep(nsec);
				fl_break();
			}
			else if (fl->verbose > 1)
				fprintf(stderr,
					"set DEBUG_FILTER to debug the child in gdb\n");
		}
#endif

		if (read_fname(fl))
			rtc = 1;
		else
		{
			if ((fl->data_fp = fopen(fl->data_fname, "r")) != NULL)
			{
				static const char infomsg[] = "INFO";
				char const *logmsg = infomsg;
				
				/* kill me after 15 minutes, no matter what */
				fl_reset_signal();
				alarm(900);

				fl->resp = NULL;
				if (fl->filter_fn)
					(*fl->filter_fn)(fl);
				resp = fl->resp;
				
				alarm(0);

				fclose(fl->data_fp);
				fl->data_fp = NULL;
				
				while (fl->cfc)
					free(cfc_shift(&fl->cfc));
				if (!fl_keep_running() || resp == NULL)
					logmsg = "ERR";
				if (fl->verbose > 3 || fl->verbose && logmsg != infomsg)
					fprintf(stderr, "%s:"
						THE_FILTER "[%d]: %s%s", logmsg, (int)getpid(),
						fl_keep_running() ? "" : "interrupted: ",
						resp ? resp : "(null response)");
			}
			else
			{
				fprintf(stderr, "ALERT:"
					THE_FILTER ": cannot open %s: %s\n",
					fl->data_fname, strerror(errno));
			}
			free(fl->data_fname);
			fl->data_fname = NULL;
		}
		
		/* 
		** give response
		*/
		if (rtc == 0 && resp == NULL)
		{
			resp = "432 Mail filter temporarily unavailable.\n";
			fprintf(stderr, "ERR:"
				THE_FILTER "[%d]: response was NULL!!\n",
				(int)getpid());
		}
		l = rtc ? 0 : strlen(resp);
		while (w < l && fl_keep_running())
		{
			unsigned p = write(fd_out, resp + w, l - w);
			if (p == (unsigned)(-1))
			{
				switch (errno)
				{
					case EAGAIN:
					case EINTR:
						continue;
					default:
						break;
				}
				fprintf(stderr, "ALERT:"
					THE_FILTER "[%d]: unable to write resp: %s\n",
					(int)getpid(), strerror(errno));
				break;
			}
			w += p;
		}
		exit(rtc);
	}
	else if (pid > 0) /* parent */
		++live_children;
	else
		perror("ALERT:" THE_FILTER ": fork");
}

static int my_lf_accept(int listensock)
/*
** copied from courier/filters/libfilter/libfilter.c
** changed: different return code for shutting down (0 instead of -1)
*/
{
	struct sockaddr_un ssun;
	fd_set fd0;
	int fd;
	int sunlen;

	if (listensock <= 0)
		return 0;
	
	for (;;)
	{
		FD_ZERO(&fd0);
		FD_SET(0, &fd0);
		FD_SET(listensock, &fd0);
 
		if (select(listensock+1, &fd0, 0, 0, 0) < 0)
		{
			return -1;
		}

		if (FD_ISSET(0, &fd0))
		{
			char buf[16];

			if (read(0, buf, sizeof(buf)) <= 0)
				return 0; /* 0 is Shutting down (cannot be accepted socket) */
		}

		if (!FD_ISSET(listensock, &fd0))
			continue;

		sunlen = sizeof ssun;
		if ((fd = accept(listensock, (struct sockaddr*)&ssun, &sunlen)) < 0)
			continue;

		fcntl(fd, F_SETFL, 0);  /* Take out of NDELAY mode */
		break;
	}
	
	return fd;
}

/* ----- init, test, main functions ----- */

static void fl_init(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof act);
	sigemptyset(&act.sa_mask);
		
	act.sa_flags = SA_NOCLDSTOP;
	act.sa_handler = child_reaper;
	sigaction(SIGCHLD, &act, NULL);
		
	act.sa_flags = 0;
	act.sa_handler = sig_catcher;
	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGHUP, &act, NULL);
	sigaction(SIGUSR1, &act, NULL);
	sigaction(SIGUSR2, &act, NULL);
}

static int fl_runtest(fl_parm* fl, int argc, char *argv[])
{
	int rtc = 0;
	int mypipe[2], i;
	if (pipe(mypipe) == 0)
	{
		FILE *fp = fdopen(mypipe[1], "w");
		for (i = 0; i < argc; ++i)
		{
			fl->in = mypipe[0];
			fl->out = 1; // stdout
			fl_runchild(fl);
#if !defined(NDEBUG)
			if (fl->verbose > 1)
				fprintf(stderr, THE_FILTER "[%d]: Running %d child(ren)\n",
					(int)getpid(), live_children);
#endif
			fprintf(fp, "%s\nthe ids\n\n", argv[i]);
			fflush(fp);
		}
		fclose(fp);
		close(mypipe[0]);
		sleep(1);
	}
	else
	{
		rtc = 1;
		perror("cannot pipe");
	}
	return rtc;
}

static void
run_sig_function(fl_init_parm const*fn, fl_parm *fl, int sig)
{
	fl_callback handler;
	switch (sig)
	{
		case SIGHUP:
			handler = fn->on_sighup;
			break;

		case SIGUSR1:
			handler = fn->on_sigusr1;
			break;

		case SIGUSR2:
			handler = fn->on_sigusr2;
			break;

		default:
			handler = NULL;
			break;
	}

	if (handler)
		(*handler)(fl);
}

static int fl_run_batchtest(fl_init_parm const*fn, fl_parm *fl)
{
	int rtc = 0;
	int mypipe[2];
	if (pipe(mypipe) == 0)
	{
		FILE *fp = fdopen(mypipe[1], "w");
		int pending = 0, total = 0;
		unsigned sleep_arg = 0;
		
		fl->in = mypipe[0];
		fl->out = 1; // stdout

		while (!feof(stdin))
		{
			char cmdbuf[1024];
			unsigned char *s, *es;
			
			if (sleep_arg == 0)
			{
				if (fl->verbose)
				{
					fprintf(stdout, "%d: ", total);
					fflush(stdout);
				}
				errno = 0;
				s = fgets(cmdbuf, sizeof cmdbuf, stdin);
				if (signal_hangup)
				{
					int const sig = signal_hangup;
					signal_hangup = 0;
					run_sig_function(fn, fl, sig);
				}
			}
			else
				s = NULL;
			if (s || sleep_arg)
			{
				int is_exit = 0;
				if (s)
				{
					es = s + strlen(s);
					while (--es >= s && isspace(*es))
						*es = 0;
					++es;
					is_exit = strcmp(s, "exit") == 0;
				}
				
				if (s == es || is_exit || sleep_arg)
				{
					if (pending)
					{
						fputc('\n', fp);
						fflush(fp);
						fl_runchild(fl);
						pending = 0;
						++total;
					}
					if (is_exit)
						break;
					if (sleep_arg)
					{
						sleep(sleep_arg);
						sleep_arg = 0;
						if (signal_hangup)
						{
							int const sig = signal_hangup;
							signal_hangup = 0;
							run_sig_function(fn, fl, sig);
						}
					}
				}
				else if (strcmp(s, "sigusr1") == 0)
					run_sig_function(fn, fl, SIGUSR1);
				else if (strcmp(s, "sigusr2") == 0)
					run_sig_function(fn, fl, SIGUSR2);
				else if (strcmp(s, "sighup") == 0)
					run_sig_function(fn, fl, SIGHUP);
				else if (strncmp(s, "test", 4) == 0)
				{
					fl_callback handler;
					switch (s[4])
					{
						case '1': handler = fn->test_fn1; break;
						case '2': handler = fn->test_fn2; break;
						case '3': handler = fn->test_fn3; break;
						case '4': handler = fn->test_fn4; break;
						default:
							handler = NULL;
							break;
					}
					if (handler)
						(*handler)(fl);
				}
				else if (strncmp(s, "sleep ", 6) == 0)
				{
					char *t = NULL;
					unsigned l = strtoul(s + 6, &t, 0);
					if (l == 0 || t == NULL || *t != 0)
						fputs("sleep requires a number\n", stderr);
					else
						sleep_arg = l;
				}
				else if (strcmp(s, "?") == 0) fputs(
				"filterlib batch test: recognized commands are:\n"
				" sigusr1, sigusr2, sighup   - simulate the corresponding signal\n"
				" test1, test2, test3, test4 - call app function, if defined\n"
				"                              (usually test1 prints configuration)\n"
				" sleep nn                   - sleep nn seconds or until signal\n"
				" exit                       - terminate batch testing\n"
				"unrecognized lines are interpreted as mail files and passed to the\n"
				"filter: mail file first, any number of ctl files until an empty line\n"
				"or one with a recognized command\n", stdout);
				
				else
				{
					if (fl->verbose > 2)
						fprintf(stdout,
							"interpreted as %s file\n",
								pending == 0 ? "mail" : "ctl");
					fprintf(fp, "%s\n", s);
					++pending;
				}
			}
			else if (ferror(stdin))
			{
				int const handled = errno == EINTR || errno == EAGAIN;
				if (fl->verbose > 2 || !handled)
					fprintf(stderr, "error reading stdin: %s\n",
						strerror(errno));
				if (!handled)
				{
					rtc = 1;
					break;
				}
				if (!feof(stdin))
					clearerr(stdin);
			}
		}
		fclose(fp);
		close(mypipe[0]);
		fprintf(stderr, THE_FILTER ": batch run %d msg(s)\n", total);
	}
	else
	{
		rtc = 1;
		perror("cannot pipe");
	}
	return rtc;
}

static int call_lf_init(void)
{
	int listensock = lf_init("filters/" THE_FILTER "-mode",
		ALLFILTERSOCKETDIR "/" THE_FILTER,
		ALLFILTERSOCKETDIR "/." THE_FILTER,
		FILTERSOCKETDIR "/" THE_FILTER,
		FILTERSOCKETDIR "/." THE_FILTER);
		
	if (listensock < 0)
	{
		perror("ALERT:" THE_FILTER ": lf_init failed");
	}
	
	return listensock;
}

static int is_courierfilter(int verbose)
{
	int rtc = 1;
#if defined(COURIERFILTER_SETS_FD3)
	struct stat stat;
	if (fstat(3, &stat) != 0)
	{
		if (errno == EBADF)
		{
			rtc = 0;
		}
		else if (verbose)
			fprintf(stderr, THE_FILTER ": cannot fstat 3: %s\n",
				strerror(errno));
	}
	else
	{
#if !defined(NDEBUG)
		fprintf(stderr, THE_FILTER ": fd 3 has mode=%lx size=%ld fstype=%.*s\n",
			(unsigned long)stat.st_mode,
			(unsigned long)stat.st_size,
#if HAVE_ST_FSTYPE_STRING			
			(int)sizeof stat.st_fstype,
			stat.st_fstype);
#else
			6,
			"undef.");
#endif // HAVE_ST_FSTYPE_STRING
#endif // !defined(NDEBUG)

		if (stat.st_size > 1) // available to read
			rtc = 0;
	}
	
	if (rtc == 0)
		fprintf(stderr, THE_FILTER ": bad fd3: invalid call\n");
#else
	(void)verbose;
#endif // COURIERFILTER_SETS_FD3
	return rtc;
}

int fl_main(fl_init_parm const*fn, void *parm,
	int argc, char *argv[], int verbose)
{
	int rtc = 0, wait_child = 5, i;
	fl_parm fl;

	memset(&fl, 0, sizeof fl);
	fl.parm = parm;
	fl.verbose = sig_verbose = verbose;
	fl.filter_fn = fn ? fn->filter_fn : NULL;
	fl.argv0 = argv[0];
	
	for (i = 1; i < argc; ++i)
	{
		char const *const arg = argv[i];
		if (strcmp(arg, "-t") == 0)
		{
			fl.testing = 1;
			fl_init();
			if (fl.verbose > 1)
				fprintf(stderr,
					THE_FILTER ": running test on %d files\n",
					argc - i - 1);
			rtc = fl_runtest(&fl, argc - i - 1, argv + i + 1);
			break;
		}

		if (strcmp(arg, "--batch-test") == 0)
		{
			fl.batch_test = fl.testing = 1;
			fl_init();
			if (fl.verbose)
				fprintf(stdout,
					THE_FILTER ": batch test. Type `?' for help.\n");
			rtc = fl_run_batchtest(fn, &fl);
			break;
		}
		
		if (strcmp(arg, "--help") == 0)
		{
			fputs(
			/*  12345678901234567890123456 */
				"  -t file...              scan rest of args as mail file(s)\n"
				"  --batch-test            enter batch test mode\n",
					stdout);
			return 1;
		}
	}

	if (fl.testing == 0 &&
		argc == 1 &&
		is_courierfilter(fl.verbose)) /* install filter */
	{
		int listensock = -1;
		
		if (fl.verbose > 2)
			fprintf(stderr,
				THE_FILTER ": running\n");
		setsid();
		/*
		int rtc = setpgrp();
	
		if (rtc)
			fprintf(stderr, THE_FILTER ": cannot set process group\n");
		*/

		fl_init();
		listensock = call_lf_init();

		if (listensock < 0)
			return 1;
		
		if (fn->init_complete)
			(*fn->init_complete)(&fl);

		lf_init_completed(listensock);

		for (;;)
		{
			int fd;
			int const sig = signal_hangup;

			if (sig != 0)
			{
				signal_hangup = 0;
				run_sig_function(fn, &fl, sig);
			}
			
			if ((fd = my_lf_accept(listensock)) <= 0)
			{
				if (fd < 0) /* select interrupted */
				{
					switch (errno)
					{
						case EAGAIN:
						case EINTR:
							continue;
						default:
							break;
					}
					perror("select");
					rtc = 1;
				}
				/* fd == 0 for clean shutdown */
				break;
			}
			signal_timed_out = signal_break = 0;
			fl.in = fl.out = fd;
			fl_runchild(&fl);
			close(fd);
		}		
	}

	if (fl.verbose && live_children == 0)
		fprintf(stderr, THE_FILTER "[%d]: exiting\n", (int)getpid());

	wait_child += live_children;

#if !defined(NDEBUG)
	i = getenv("DEBUG_FILTER") ? 1 : 0;
	if (i == 1 && live_children == 1 &&
		fl_get_test_mode(&fl) == fl_testing)
	{
		fprintf(stderr, THE_FILTER
			": leaving the child running for dbg\n");
		return rtc;
	}
#else
	i = 0;
#endif

	while (live_children > 0)
	{
		int nsec =  live_children*3;

		if (fl.batch_test == 0 && i == 0)
			fprintf(stderr, THE_FILTER "[%d]: waiting for %d child(ren)\n",
				(int)getpid(), live_children);

		if (sleep(nsec) != 0)
			continue;

		if (fl.testing == 0)
		{
			kill(0, SIGTERM);
			if (--wait_child < 0)
			{
				fprintf(stderr, "WARN:"
					THE_FILTER
					"[%d]: leaving %d naughty child(ren) running\n",
					(int)getpid(), live_children);
			}
		}
	}
	
	return rtc;
}
